from django.forms import ModelForm
from edc_crf.modelform_mixins import CrfModelFormMixin

from edc_visit_tracking.form_validators import VisitMissedFormValidator
from edc_visit_tracking.models import SubjectVisitMissed


class SubjectVisitMissedForm(CrfModelFormMixin, ModelForm):
    form_validator_cls = VisitMissedFormValidator

    def validate_against_consent(self):
        pass

    class Meta:
        model = SubjectVisitMissed
        fields = "__all__"



from datetime import datetime
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_appointment.constants import INCOMPLETE_APPT
from edc_appointment.managers import AppointmentDeleteError
from edc_appointment.models import Appointment
from edc_appointment.utils import reset_appointment, skip_appointment
from edc_consent import site_consents
from edc_facility.import_holidays import import_holidays
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED
from edc_visit_tracking.model_mixins import PreviousVisitError
from edc_visit_tracking.models import SubjectVisit
from edc_visit_tracking.visit_sequence import VisitSequence, VisitSequenceError
from visit_tracking_app.consents import consent_v1
from visit_tracking_app.visit_schedule import visit_schedule1, visit_schedule2

from ..helper import Helper

utc_tz = ZoneInfo("UTC")


class DisabledVisitSequence(VisitSequence):
    def enforce_sequence(self, **kwargs) -> None:
        return None


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
@override_settings(SUBJECT_SCREENING_MODEL="visit_tracking_app.subjectscreening")
class TestPreviousVisit(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        site_consents.registry = {}
        site_consents.register(consent_v1)
        SubjectVisit.visit_sequence_cls = VisitSequence
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name=visit_schedule1.name, schedule_name="schedule1"
        )

    def test_visit_sequence_enforcer_on_first_visit_in_sequence(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.visit_sequence_cls = DisabledVisitSequence
        visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        visit_sequence = VisitSequence(appointment=visit.appointment)
        try:
            visit_sequence.enforce_sequence()
        except VisitSequenceError as e:
            self.fail(f"VisitSequenceError unexpectedly raised. Got '{e}'")

    def test_visit_sequence_enforcer_without_first_visit_in_sequence(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.visit_sequence_cls = DisabledVisitSequence
        visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        visit_sequence = VisitSequence(appointment=visit.appointment)
        self.assertRaises(VisitSequenceError, visit_sequence.enforce_sequence)

    def test_requires_previous_visit_thru_model(self):
        """Asserts requires previous visit to exist on create."""
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[2],
            report_datetime=get_utcnow() - relativedelta(months=8),
            reason=SCHEDULED,
        )
        SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[3],
            report_datetime=get_utcnow() - relativedelta(months=8),
            reason=SCHEDULED,
        )

    def test_requires_previous_visit_thru_model2(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )

        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[2],
            report_datetime=get_utcnow() - relativedelta(months=8),
        )

    def test_previous_appointment(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        visit_sequence = VisitSequence(appointment=appointments[0], skip_enforce=True)
        self.assertIsNone(visit_sequence.previous_appointment)
        visit_sequence = VisitSequence(appointment=appointments[1], skip_enforce=True)
        self.assertEqual(visit_sequence.previous_appointment, appointments[0])
        visit_sequence = VisitSequence(appointment=appointments[2], skip_enforce=True)
        self.assertEqual(visit_sequence.previous_appointment, appointments[1])

    def test_previous_appointment_with_unscheduled(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()
            appointment = self.helper.create_unscheduled(appointment)
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        visit_sequence = VisitSequence(appointment=appointments[0])
        self.assertIsNone(visit_sequence.previous_appointment)
        for i in range(0, Appointment.objects.all().count() - 1):
            visit_sequence = VisitSequence(appointment=appointments[i + 1])
            self.assertEqual(visit_sequence.previous_appointment, appointments[i])

    def test_previous_appointment_broken_sequence1(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        self.assertRaises(AppointmentDeleteError, appointments[1].delete)

    def test_previous_visit_report_broken_sequence2(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()
            appointment = self.helper.create_unscheduled(appointment)
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()
            appointment = self.helper.create_unscheduled(appointment)
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        visit_sequence = VisitSequence(appointment=appointments[0])  # 1000.0
        self.assertIsNone(visit_sequence.previous_appointment)
        visit_sequence = VisitSequence(appointment=appointments[1])  # 1000.1
        self.assertEqual(visit_sequence.previous_appointment, appointments[0])  # 1000.0

        appointments[1].related_visit.delete()

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        visit_sequence = VisitSequence(appointment=appointments[2])  # 1000.2
        self.assertRaises(VisitSequenceError, visit_sequence.enforce_sequence)

        visit_sequence = VisitSequence(appointment=appointments[3])
        self.assertRaises(VisitSequenceError, getattr, visit_sequence, "previous_appointment")

    def test_previous_visit(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED,
            )

    def test_previous_visit_with_inserted_unscheduled(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()
            appointment = self.helper.create_unscheduled(appointment)
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()

    def test_requires_previous_visit_unless_skipped(self):
        """Asserts does not require previous visit if previous appt
        is skipped.
        """
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )

        skip_appointment(appointments[1])

        try:
            SubjectVisit.objects.create(
                appointment=appointments[2],
                report_datetime=get_utcnow() - relativedelta(months=8),
                reason=SCHEDULED,
            )
        except PreviousVisitError:
            self.fail("PreviousVisitError unexpectedly raised.")

        reset_appointment(appointments[1])

        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[2],
            report_datetime=get_utcnow() - relativedelta(months=8),
            reason=SCHEDULED,
        )



from __future__ import annotations

from datetime import datetime
from typing import Any, Generic, Protocol, Tuple, Type, TypeVar
from uuid import UUID

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.sites.models import Site
from django.db.models import ForeignObjectRel, Model
from django.db.models.manager import BaseManager
from django.forms import Field
from edc_appointment.models import Appointment
from edc_metadata.metadata import Destroyer, Metadata
from edc_metadata.metadata_rules import MetadataRuleEvaluator
from edc_visit_schedule.schedule import Schedule, VisitCollection
from edc_visit_schedule.typing_stubs import VisitScheduleFieldsProtocol
from edc_visit_schedule.visit import Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

_M = TypeVar("_M", bound=Model)
_Self = TypeVar("_Self", bound=Model)


class ModelBase(type):
    @property
    def objects(cls: type[_Self]) -> BaseManager[_Self]: ...  # noqa

    @property
    def _default_manager(cls: type[_Self]) -> BaseManager[_Self]: ...  # noqa

    @property
    def _base_manager(cls: type[_Self]) -> BaseManager[_Self]: ...  # noqa


class Options(Generic[_M]):
    def label_lower(self) -> str: ...  # noqa

    def fields(self) -> Tuple[Field]: ...  # noqa

    def get_fields(
        self, include_parents: bool = ..., include_hidden: bool = ...  # noqa
    ) -> list[Field | ForeignObjectRel | GenericForeignKey]: ...  # noqa


class SiteFieldsProtocol(Protocol):
    id: int
    name: str | None


class RelatedVisitProtocol(VisitScheduleFieldsProtocol, Protocol):
    metadata_cls: Type[Metadata]
    metadata_destroyer_cls: Type[Destroyer]
    metadata_rule_evaluator_cls: Type[MetadataRuleEvaluator]

    class Meta: ...  # noqa

    _meta: Options[Any]

    appointment: Appointment
    consent_version: str | None
    created: datetime | None
    id: UUID
    modified: datetime | None
    reason: str | None
    report_datetime: datetime | None
    schedule: Schedule
    site: Site
    subject_identifier: str | None
    user_created: str | None
    user_modified: str | None
    visit: Visit
    visit_schedule: VisitSchedule
    visits: VisitCollection



from django.contrib import admin
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin

from .admin_site import edc_visit_tracking_admin
from .modeladmin_mixins import VisitModelAdminMixin
from .models import SubjectVisit


@admin.register(SubjectVisit, site=edc_visit_tracking_admin)
class SubjectVisitAdmin(
    SiteModelAdminMixin,
    VisitModelAdminMixin,
    ModelAdminSubjectDashboardMixin,
    SimpleHistoryAdmin,
):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Type
from uuid import UUID

from edc_consent.fieldsets import REQUIRES_CONSENT_FIELDS
from edc_constants.constants import UUID_PATTERN

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from edc_crf.model_mixins import CrfModelMixin

    from ..model_mixins import VisitModelMixin


class CrfModelAdminMixin:
    """ModelAdmin subclass for models with a ForeignKey to your
    visit model(s).
    """

    date_hierarchy: str = "report_datetime"
    report_datetime_field_attr: str = "report_datetime"

    def visit_reason(self, obj: CrfModelMixin | None = None) -> str:
        return getattr(obj, self.related_visit_model_attr).reason

    def visit_code(self, obj: CrfModelMixin | None = None) -> str:
        return getattr(obj, self.related_visit_model_attr).visit_code

    def visit_code_sequence(self, obj: CrfModelMixin | None = None) -> int:
        return getattr(obj, self.related_visit_model_attr).visit_code_sequence

    def subject_identifier(self, obj: CrfModelMixin | None = None) -> str:
        return getattr(obj, self.related_visit_model_attr).subject_identifier

    def get_list_display(self, request: WSGIRequest) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        fields_first: tuple[str, str, str, str] = (
            "subject_identifier",
            self.report_datetime_field_attr,
            "visit_code",
            "visit_reason",
        )
        return fields_first + tuple(
            f for f in list_display if f not in fields_first + ("__str__",)
        )

    def get_search_fields(self, request: WSGIRequest) -> tuple[str, ...]:
        search_fields = super().get_search_fields(request)
        field = (
            f"{self.related_visit_model_attr}__appointment__subject_identifier",
            f"{self.related_visit_model_attr}__visit_code",
        )
        if field not in search_fields:
            return search_fields + field
        return search_fields

    def get_list_filter(self, request: WSGIRequest) -> tuple[str, ...]:
        """Returns a tuple of list_filters.

        Not working?? Call `get_list_filter`, don't explicitly set `list_filter`
        in the concrete class or any of the mixins.
        """
        list_filter = super().get_list_filter(request)
        fields = (
            f"{self.related_visit_model_attr}__{self.report_datetime_field_attr}",
            f"{self.related_visit_model_attr}__visit_code",
            f"{self.related_visit_model_attr}__visit_code_sequence",
            f"{self.related_visit_model_attr}__reason",
        )
        return tuple(f for f in list_filter if f not in fields) + fields

    @property
    def related_visit_model_attr(self) -> str:
        return self.model.related_visit_model_attr()

    @property
    def related_visit_model_cls(self) -> Type[VisitModelMixin]:
        return self.model.related_visit_model_cls()

    def related_visit(self, request: WSGIRequest, obj=None) -> VisitModelMixin | None:
        """Returns the related_visit from the request object

        You may need to wrap this in an exception in some
        cases.
        """
        related_visit = self.related_visit_model_cls.objects.get(
            id=self.related_visit_id(request)
        )
        if not related_visit and obj:
            related_visit = getattr(obj, self.related_visit_model_attr)
        return related_visit

    def related_visit_id(self, request) -> UUID | str | None:
        """Returns the record id/pk for the related visit"""
        related_visit_id = request.GET.get(self.related_visit_model_attr)
        if related_visit_id and UUID_PATTERN.match(related_visit_id):
            return related_visit_id
        return None

    def formfield_for_foreignkey(self, db_field, request: WSGIRequest, **kwargs):
        db = kwargs.get("using")
        if db_field.name == self.related_visit_model_attr:
            related_visit_id = self.related_visit_id(request)
            if related_visit_id and UUID_PATTERN.match(related_visit_id):
                kwargs["queryset"] = self.related_visit_model_cls._default_manager.using(
                    db
                ).filter(id__exact=related_visit_id)
            else:
                kwargs["queryset"] = self.related_visit_model_cls._default_manager.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def get_readonly_fields(
        self, request: WSGIRequest, obj: CrfModelMixin | None = None
    ) -> tuple[str]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        readonly_fields += REQUIRES_CONSENT_FIELDS
        if (
            not self.related_visit_id(request)
            and self.related_visit_model_attr not in readonly_fields
        ):
            readonly_fields += (self.related_visit_model_attr,)
        return readonly_fields



from __future__ import annotations

from typing import TYPE_CHECKING, Type

from django.conf import settings
from django.contrib.admin.utils import NotRelationField, get_model_from_relation
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.db import models

from ...exceptions import RelatedVisitFieldError
from ..visit_model_mixin import VisitModelMixin

if TYPE_CHECKING:
    from django.db.models import OneToOneField


class VisitMethodsModelMixin(models.Model):
    """A model mixin for CRFs and Requisitions to add methods to
    access the related visit model and its attributes.

    Used by VisitTrackingCrfModelMixin.
    """

    def __str__(self) -> str:
        return str(self.related_visit)

    def natural_key(self) -> tuple:
        return tuple(getattr(self, self.related_visit_model_attr()).natural_key())

    natural_key.dependencies = [settings.SUBJECT_VISIT_MODEL]

    @classmethod
    def related_visit_model_attr(cls) -> str:
        """Returns the field name for the related visit model
        foreign key.
        """
        return cls.related_visit_field_cls().name
        # return get_related_visit_model_attr(cls)

    @classmethod
    def related_visit_field_cls(cls) -> OneToOneField | None:
        """Returns the 'field' class of the related visit foreign
        key attribute.
        """
        related_visit_field_cls = None
        for field in cls._meta.get_fields():
            try:
                related_model = get_model_from_relation(field)
            except NotRelationField:
                continue
            else:
                if issubclass(related_model, (VisitModelMixin,)):
                    related_visit_field_cls = field
                    break
        if not related_visit_field_cls:
            raise RelatedVisitFieldError(f"Related visit field class not found. See {cls}.")
        return related_visit_field_cls

    @classmethod
    def related_visit_model_cls(cls) -> Type[VisitModelMixin]:
        """Returns the 'model' class of the related visit foreign
        key attribute.
        """
        related_model = None
        for field in cls._meta.get_fields():
            try:
                related_model = get_model_from_relation(field)
            except NotRelationField:
                continue
            else:
                if issubclass(related_model, (VisitModelMixin,)):
                    break
        if not related_model:
            raise RelatedVisitFieldError(f"Related visit field class not found. See {cls}.")
        return related_model

    @classmethod
    def related_visit_model(cls) -> str:
        """Returns the name of the related_visit FK model in
        label_lower format.
        """
        return cls.related_visit_model_cls()._meta.label_lower

    @property
    def visit_code(self) -> str:
        return self.related_visit.visit_code

    @property
    def visit_code_sequence(self) -> int:
        return self.related_visit.visit_code_sequence

    @property
    def related_visit(self) -> VisitModelMixin:
        """Returns the instance of the related_visit FK."""
        related_model = None
        related_visit = None
        for field in self._meta.get_fields():
            try:
                related_model = get_model_from_relation(field)
            except NotRelationField:
                continue
            else:
                if issubclass(related_model, (VisitModelMixin,)):
                    try:
                        related_visit = getattr(self, field.name)
                    except ObjectDoesNotExist:
                        raise RelatedVisitFieldError(
                            f"Related visit cannot be None. See {self.__class__}. "
                            "Perhaps catch this in the form."
                        )
                    break
                else:
                    related_model = None
        if not related_model:
            raise ImproperlyConfigured(
                f"Model is missing a FK to a related visit model. See {self.__class__}."
            )
        return related_visit

    class Meta:
        abstract = True



from datetime import datetime
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, override_settings
from edc_appointment.constants import MISSED_APPT, ONTIME_APPT, SCHEDULED_APPT
from edc_appointment.exceptions import AppointmentBaselineError
from edc_appointment.models import Appointment
from edc_consent import site_consents
from edc_constants.constants import (
    ALIVE,
    DEAD,
    HOSPITALIZED,
    NO,
    NOT_APPLICABLE,
    OTHER,
    YES,
)
from edc_facility.import_holidays import import_holidays
from edc_list_data import load_list_data
from edc_metadata.models import CrfMetadata
from edc_utils import get_utcnow
from edc_visit_schedule.constants import DAY1
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.visit import Crf, CrfCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from edc_visit_tracking.constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from edc_visit_tracking.models import SubjectVisit, SubjectVisitMissedReasons
from visit_tracking_app.consents import consent_v1
from visit_tracking_app.models import list_data

from ..forms import SubjectVisitMissedForm
from ..helper import Helper

utc_tz = ZoneInfo("UTC")


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestVisit(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        load_list_data(
            list_data=list_data,
            model_name="edc_visit_tracking.subjectvisitmissedreasons",
        )
        self.subject_identifier = "12345"
        site_consents.registry = {}
        site_consents.register(consent_v1)
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        crfs = CrfCollection(
            Crf(show_order=1, model="visit_tracking_app.crfone", required=True),
            Crf(show_order=2, model="visit_tracking_app.crftwo", required=True),
            Crf(show_order=3, model="visit_tracking_app.crfthree", required=True),
            Crf(show_order=4, model="visit_tracking_app.crffour", required=True),
            Crf(show_order=5, model="visit_tracking_app.crffive", required=True),
        )
        crfs_missed = CrfCollection(
            Crf(
                show_order=1,
                model="edc_visit_tracking.subjectvisitmissed",
                required=True,
            ),
        )

        visit_schedule1 = VisitSchedule(
            name="visit_schedule1",
            offstudy_model="visit_tracking_app.subjectoffstudy",
            death_report_model="visit_tracking_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule1 = Schedule(
            name="schedule1",
            onschedule_model="visit_tracking_app.onscheduleone",
            offschedule_model="visit_tracking_app.offscheduleone",
            consent_definitions=[consent_v1],
        )
        visits = []
        for index in range(0, 4):
            visits.append(
                Visit(
                    code=f"{index + 1}000",
                    title=f"Day {index + 1}",
                    timepoint=index,
                    rbase=relativedelta(days=index),
                    rlower=relativedelta(days=0),
                    rupper=relativedelta(days=6),
                    requisitions=None,
                    crfs=crfs,
                    crfs_missed=crfs_missed,
                    allow_unscheduled=True,
                )
            )
        for visit in visits:
            schedule1.add_visit(visit)
        visit_schedule1.add_schedule(schedule1)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)

    @staticmethod
    def get_subject_visit(
        visit_code=None,
        visit_code_sequence=None,
        appt_timing=None,
        visit_reason=None,
    ):
        appt_timing = appt_timing or ONTIME_APPT
        if visit_reason == MISSED_VISIT:
            appt_timing = MISSED_APPT
        visit_code = visit_code or DAY1
        visit_code_sequence = 0 if visit_code_sequence is None else visit_code_sequence
        if not visit_reason:
            if visit_code_sequence > 0:
                visit_reason = UNSCHEDULED
            else:
                visit_reason = MISSED_VISIT if appt_timing == MISSED_APPT else SCHEDULED_APPT
        appointment = Appointment.objects.get(
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
        )
        appointment.appt_timing = appt_timing
        appointment.save()
        opts = dict(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=visit_reason,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
        )
        try:
            subject_visit = SubjectVisit.objects.get(appointment=appointment)
        except ObjectDoesNotExist:
            subject_visit = SubjectVisit.objects.create(**opts)
        else:
            for k, v in opts.items():
                setattr(subject_visit, k, v)
            subject_visit.save()
            subject_visit.refresh_from_db()
        return appointment, subject_visit

    @override_settings(
        SUBJECT_MISSED_VISIT_REASONS_MODEL="edc_visit_tracking.subjectvisitmissed"
    )
    def test_(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        # baseline
        appointment, _ = self.get_subject_visit(appt_timing=ONTIME_APPT)
        appointment = appointment.next
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()
        appointment, subject_visit = self.get_subject_visit(
            appt_timing=MISSED_APPT, visit_code=appointment.visit_code
        )
        opts = dict(
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            timepoint=appointment.timepoint,
        )
        self.assertGreater(CrfMetadata.objects.filter(**opts).count(), 0)
        subject_visit.reason = MISSED_VISIT
        subject_visit.save()
        self.assertEqual(1, CrfMetadata.objects.filter(**opts).count())

    @override_settings(
        SUBJECT_MISSED_VISIT_REASONS_MODEL="edc_visit_tracking.subjectvisitmissed"
    )
    def test_baseline_appt_can_never_be_missed(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[0]
        appointment.appt_timing = MISSED_APPT
        self.assertRaises(AppointmentBaselineError, appointment.save)

    @override_settings(
        SUBJECT_MISSED_VISIT_REASONS_MODEL="edc_visit_tracking.subjectvisitmissed"
    )
    def test_baseline_visit_can_never_be_missed(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertRaises(
            AppointmentBaselineError, self.get_subject_visit, visit_reason=MISSED_VISIT
        )

    @override_settings(
        SUBJECT_MISSED_VISIT_REASONS_MODEL="edc_visit_tracking.subjectvisitmissed"
    )
    def test_missed_appt_updates_subject_visit_as_missed(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        appointment, _ = self.get_subject_visit()
        appointment = appointment.next
        appointment.appt_timing = ONTIME_APPT
        appointment, subject_visit = self.get_subject_visit(
            visit_code=appointment.visit_code, visit_reason=SCHEDULED
        )
        self.assertEqual(subject_visit.reason, SCHEDULED)
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()
        self.assertEqual(appointment.visit_code, "2000")
        subject_visit.refresh_from_db()
        self.assertEqual(subject_visit.reason, MISSED_VISIT)
        self.assertEqual(subject_visit.visit_code, "2000")

    @override_settings(
        SUBJECT_MISSED_VISIT_REASONS_MODEL="edc_visit_tracking.subjectvisitmissed"
    )
    def test_missed_appt_updates_subject_visit_as_missed2(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
            report_datetime=get_utcnow() - relativedelta(months=6),
        )
        appointment, _ = self.get_subject_visit()
        appointment = appointment.next
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment, subject_visit = self.get_subject_visit(
            visit_code=appointment.visit_code, visit_reason=MISSED_VISIT
        )
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            contact_last_date=subject_visit.report_datetime,
            survival_status=ALIVE,
            contact_attempted=YES,
            contact_made=YES,
            contact_attempts_count=3,
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            ltfu=NO,
            site=Site.objects.get(id=settings.SITE_ID),
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

        form.save(commit=True)

    def test_subject_visit_missed_form_survivial_and_ltfu(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        _, subject_visit = self.get_subject_visit()
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=DEAD,
            contact_attempted=YES,
            contact_attempts_count=4,
            contact_attempts_explained=None,
            contact_last_date=get_utcnow(),
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            contact_made=YES,
            ltfu=YES,
            site=Site.objects.get(id=settings.SITE_ID),
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("ltfu", form._errors)

        data.update(survival_status=DEAD, ltfu=NOT_APPLICABLE)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("ltfu", form._errors)

        data.update(survival_status=ALIVE, ltfu=NOT_APPLICABLE)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("ltfu", form._errors)

    def test_subject_visit_missed_form_missed_reasons(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        _, subject_visit = self.get_subject_visit()
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=ALIVE,
            contact_attempted=YES,
            contact_attempts_count=1,
            contact_last_date=get_utcnow(),
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            contact_made=YES,
            ltfu=NO,
            site=Site.objects.get(id=settings.SITE_ID),
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("missed_reasons_other", form._errors)
        data.update(missed_reasons=[SubjectVisitMissedReasons.objects.get(name=OTHER)])
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("missed_reasons_other", form._errors)

    def test_subject_visit_missed_form_attempts(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        _, subject_visit = self.get_subject_visit()
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=ALIVE,
            contact_attempted=NO,
            contact_made=NOT_APPLICABLE,
            contact_attempts_count=None,
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            ltfu=YES,
            site=Site.objects.get(id=settings.SITE_ID),
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("contact_attempts_count", form._errors)

        data.update(contact_attempted=YES, contact_made=NO, contact_attempts_count=None)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()

        self.assertIn("contact_attempts_count", form._errors)

        data.update(contact_attempted=YES, contact_made=NO, contact_attempts_count=2)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertIn("contact_attempts_explained", form._errors)

        data.update(contact_attempted=YES, contact_made=NO, contact_attempts_count=3)
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()
        self.assertNotIn("contact_attempts_explained", form._errors)

    def test_baseline_subject_visit_cannot_be_missed(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertRaises(
            AppointmentBaselineError, self.get_subject_visit, visit_reason=MISSED_VISIT
        )

    def test_subject_visit_missed_form(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        _, subject_visit = self.get_subject_visit()
        appointment = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")[1]
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        _, subject_visit = self.get_subject_visit(
            visit_code=appointment.visit_code,
            visit_code_sequence=appointment.visit_code_sequence,
            appt_timing=MISSED_APPT,
        )

        self.assertEqual(subject_visit.reason, MISSED_VISIT)
        self.assertEqual(subject_visit.appointment.appt_timing, MISSED_APPT)

        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            survival_status=ALIVE,
            contact_attempted=NO,
            contact_made=NOT_APPLICABLE,
            contact_attempts_count=None,
            missed_reasons=[SubjectVisitMissedReasons.objects.get(name=HOSPITALIZED)],
            ltfu=YES,
            site=Site.objects.get(id=settings.SITE_ID),
        )
        form = SubjectVisitMissedForm(data=data)
        form.is_valid()



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.sites.managers import CurrentSiteManager as DjangoCurrentSiteManager
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, models, transaction
from edc_constants.constants import INCOMPLETE, NOT_APPLICABLE, OTHER

from .constants import MISSED_VISIT
from .exceptions import RelatedVisitReasonError

if TYPE_CHECKING:
    from edc_appointment.models import Appointment


class CrfModelManager(models.Manager):
    """A manager class for Crf models, models that have an FK to
    the visit model.
    """

    use_in_migrations = True

    def get_by_natural_key(
        self,
        subject_identifier,
        visit_schedule_name,
        schedule_name,
        visit_code,
        visit_code_sequence,
    ):
        instance = self.model.visit_model_cls().objects.get_by_natural_key(
            subject_identifier,
            visit_schedule_name,
            schedule_name,
            visit_code,
            visit_code_sequence,
        )
        return self.get(**{self.model.related_visit_model_attr(): instance})


class VisitModelManager(models.Manager):
    """A manager class for related visit models (e.g. subject_visit)."""

    use_in_migrations = True

    def get_by_natural_key(
        self,
        subject_identifier,
        visit_schedule_name,
        schedule_name,
        visit_code,
        visit_code_sequence,
    ):
        return self.get(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
        )

    def create_missed_extras(self) -> dict:
        """Extra options to use when auto-creating a visit
        model instance with reason=missed.

        See `create_missed_from_appointment`.
        """
        return {}

    def create_missed_from_appointment(
        self,
        appointment: Appointment,
        reason_missed: str | None = None,
        reason_missed_other: str | None = None,
    ):
        """Creates a subject visit model instance automatically
        for a missed appointment (appt_timing=missed).

        Raises if the visit model instance already exists and
        has reason!=missed.
        """
        try:
            subject_visit = self.get(appointment=appointment)
        except ObjectDoesNotExist:
            opts = dict(
                appointment=appointment,
                comments="[auto-created]",
                info_source=NOT_APPLICABLE,
                reason=MISSED_VISIT,
                reason_missed=reason_missed or OTHER,
                reason_missed_other=reason_missed_other or "[auto-created]",
                report_datetime=appointment.appt_datetime,
                schedule_name=appointment.schedule_name,
                subject_identifier=appointment.subject_identifier,
                survival_status=NOT_APPLICABLE,
                visit_code=appointment.visit_code,
                visit_code_sequence=appointment.visit_code_sequence,
                visit_schedule_name=appointment.visit_schedule_name,
            )
            opts.update(**self.create_missed_extras())
            try:
                with transaction.atomic():
                    obj = self.create(**opts)
            except IntegrityError:
                pass
            obj.document_status = INCOMPLETE
            obj.save(update_fields=["document_status"])
        else:
            if subject_visit.reason != MISSED_VISIT:
                raise RelatedVisitReasonError(
                    f"Subject visit already exists. Reason=`{subject_visit.reason}`"
                )


class VisitCurrentSiteManager(DjangoCurrentSiteManager, VisitModelManager):
    use_in_migrations = True

    pass


class CrfCurrentSiteManager(DjangoCurrentSiteManager, CrfModelManager):
    use_in_migrations = True

    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_appointment.tests.helper import Helper as BaseHelper

from visit_tracking_app.models import SubjectScreening

if TYPE_CHECKING:
    from edc_appointment.models import Appointment


class Helper(BaseHelper):
    @property
    def screening_model_cls(self):
        return SubjectScreening

    def create_unscheduled(self, appointment: Appointment):
        return self.add_unscheduled_appointment(appointment)



from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register
from django.core.management.color import color_style

from .system_checks import context_processors_check

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_visit_tracking"
    verbose_name = "Edc Visit Tracking"
    report_datetime_allowance: int = 30
    allow_crf_report_datetime_before_visit: bool = False
    reason_field: dict = {}

    def ready(self):
        register(context_processors_check)



from edc_constants.constants import (
    HOSPITAL_NOTES,
    IN_PERSON,
    NEXT_OF_KIN,
    NOT_APPLICABLE,
    OTHER,
    OTHER_PLEASE_SPECIFY_TEXT,
    OUTPATIENT_CARDS,
    PARTICIPANT,
    PATIENT,
    PATIENT_REPRESENTATIVE,
    TELEPHONE,
    TIMEPOINT,
)

from .constants import (
    CHART,
    COMPLETED_PROTOCOL_VISIT,
    DEFERRED_VISIT,
    LOST_VISIT,
    MISSED_VISIT,
    SCHEDULED,
    UNSCHEDULED,
)

ASSESSMENT_WHO_CHOICES = (
    (PATIENT, "Patient"),
    (NEXT_OF_KIN, "Next of kin"),
    (NOT_APPLICABLE, "Not applicable (if missed)"),
    (OTHER, OTHER_PLEASE_SPECIFY_TEXT),
)

ASSESSMENT_TYPES = (
    (TELEPHONE, "Telephone"),
    (IN_PERSON, "In person"),
    (NOT_APPLICABLE, "Not applicable (if missed)"),
    (OTHER, OTHER_PLEASE_SPECIFY_TEXT),
)

VISIT_REASON = (
    (SCHEDULED, "Scheduled visit/contact"),
    (UNSCHEDULED, "Unscheduled visit/contact"),
    (MISSED_VISIT, "Missed visit"),
    (LOST_VISIT, "Lost to follow-up (use only when taking subject off study)"),
    (DEFERRED_VISIT, "Deferred"),
    (COMPLETED_PROTOCOL_VISIT, "Completed protocol"),
)

VISIT_INFO_SOURCE = (
    (PARTICIPANT, "1. Clinic visit with participant"),
    ("other_contact", "2. Other contact with participant"),
    ("other_doctor", "3. Contact with external health care provider/medical doctor"),
    (
        "family",
        "4. Contact with family or designated person who can provide information",
    ),
    (CHART, "5. Hospital chart or other medical record"),
    (OTHER, "9. Other"),
)

# another option for VISIT_INFO_SOURCE
VISIT_INFO_SOURCE2 = (
    (PATIENT, "Patient"),
    (
        PATIENT_REPRESENTATIVE,
        "Patient representative (e.g., next of kin, relative, guardian)",
    ),
    (HOSPITAL_NOTES, "Hospital notes"),
    (OUTPATIENT_CARDS, "Outpatient cards"),
    (NOT_APPLICABLE, "Not applicable (if missed)"),
    (OTHER, "Other"),
)

# these defaults are not intended for production
VISIT_REASON_UNSCHEDULED = (
    ("patient_unwell_outpatient", "Patient unwell (outpatient)"),
    ("patient_hospitalised", "Patient hospitalised"),
    (OTHER, "Other"),
    (NOT_APPLICABLE, "Not applicable"),
)

# these defaults are not intended for production
VISIT_REASON_MISSED = (
    (TIMEPOINT, "Missed timepoint"),
    (OTHER, "Other"),
    (NOT_APPLICABLE, "Not applicable"),
)
